/*  Copyright (C) 1998-2002 Jason Lango
 *
 *  This file is part of the Doors for Linux Kernel Module.
 * 
 *  The Doors for Linux Kernel Module is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  The Doors for Linux Kernel Module is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE.  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Doors for Linux Kernel Module; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 *  Boston, MA  02111-1307  USA
 */

/*
 *  devdoor.c - Device interface for modularizing the Doors API.
 *  Jason Lango <jal@rampant.org>
 */

#include <linux/autoconf.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include "doorsk.h"
#include "door_ioctl.h"
#include "doors_debug.h"

MODULE_DESCRIPTION("Solaris Doors API");
MODULE_AUTHOR("Jason Lango <jal@rampant.org>");

static int devdoor_ioctl(struct inode *, struct file *,
	unsigned int, unsigned long);
static int devdoor_open(struct inode *ino, struct file *filp);
static int devdoor_release(struct inode *ino, struct file *filp);

static struct file_operations devdoor_fops = {
	ioctl:		devdoor_ioctl,
	open:		devdoor_open,
	release:	devdoor_release,
};

int devdoor_init(void)
{
	int error;

	error = door_init();
	if (error)
		goto out;

	if (register_chrdev(DOORS_MAJOR, DOORS_DEVNAME, &devdoor_fops)) {
		printk("%s: failed to register character "
			"major %d.\n", doors_banner, DOORS_MAJOR);
		return -EINVAL;
	}

out:
	return error;
}

static int devdoor_create(unsigned long arg)
{
	struct door_create_args args;

	if (copy_from_user(&args, (void*)arg, sizeof(args)))
		return -EFAULT;

	return do_door_create(args.server_procedure, args.cookie,
		args.attributes);
}

static int devdoor_fattach(unsigned long arg)
{
	struct door_fattach_args args;
	char *name;
	int error;

	if (copy_from_user(&args, (void*)arg, sizeof(args)))
		return -EFAULT;

	name = getname((void*)args.name);
	if (IS_ERR(name))
		return PTR_ERR(name);

	error = do_door_fattach(args.fd, name);
	putname(name);
	return error;
}

static int devdoor_fdetach(unsigned long arg)
{
	struct door_fdetach_args args;
	char *name;
	int error;

	if (copy_from_user(&args, (void*)arg, sizeof(args)))
		return -EFAULT;

	name = getname((void*)args.name);
	if (IS_ERR(name)) {
		return PTR_ERR(name);
	}

	error = do_door_fdetach(name);
	putname(name);
	return error;
}

static int devdoor_wait_svr_create(unsigned long arg)
{
	door_info_t info;
	int error;

	error = do_door_wait_svr_create(&info);
	if (!error) {
		if (copy_to_user((void*)arg, (void*)&info, sizeof(door_info_t)))
			return -EFAULT;
	}

	return error;
}

static int devdoor_call(unsigned long arg)
{
	struct door_call_args args;
	int error;

	if (copy_from_user(&args, (void*)arg, sizeof(args)))
		return -EFAULT;

	error = do_door_call(args.fd, &args.params);
	if (error)
		return error;

	if (copy_to_user((void*) args.rparams, &args.params,
		sizeof(door_karg_t)))
		return -EFAULT;
	return 0;
}

static int devdoor_return(unsigned long arg)
{
	struct door_return_args args;
	struct door_return_outargs outargs;
	int error;

	error = -EFAULT;
	if (copy_from_user(&args, (void*)arg, sizeof(args)))
		goto out;

	error = do_door_return((void*) args.data_ptr, args.data_size,
		(door_desc_t*) args.desc_ptr, args.num_desc,
		(void*) args.argstash, args.argstash_size,
		(unsigned long) args.munmap_addr,
		(unsigned long) args.munmap_len,
		&outargs);
	if (error)
		goto out;

	error = -EFAULT;
	if (copy_to_user((void*) args.outargs, &outargs, sizeof(outargs)))
		goto out;

	error = 0;

out:
	return error;
}

static int devdoor_info(unsigned long arg)
{
	struct door_info_args args;
	struct door_info info;
	int error;

	if (copy_from_user(&args, (void*)arg, sizeof(args)))
		return -EFAULT;

	error = do_door_info(args.fd, &info);
	if (error)
		return error;

	if (copy_to_user((void*)args.info, &info, sizeof(door_info_t)))
		return -EFAULT;
	return 0;
}

static int devdoor_ioctl(struct inode *ino, struct file *fil,
	unsigned int msg, unsigned long arg)
{
	switch (msg) {
	case DOOR_FATTACH:
		return devdoor_fattach(arg);
	case DOOR_FDETACH:
		return devdoor_fdetach(arg);
	case DOOR_CREATE:
		return devdoor_create(arg);
	case DOOR_WAIT_SVR_CREATE:
		return devdoor_wait_svr_create(arg);
	case DOOR_CALL:
		return devdoor_call(arg);
	case DOOR_RETURN:
		return devdoor_return(arg);
	case DOOR_INFO:
		return devdoor_info(arg);

	default:
		printk("%s: unknown ioctl(%u,%lu)\n", doors_banner, msg, arg);
	}
	return -ENOSYS;
}

static int devdoor_open(struct inode *ino, struct file *filp)
{
	MOD_INC_USE_COUNT;
	return 0;
}

static int devdoor_release(struct inode *ino, struct file *filp)
{
	int error;

	error = do_door_shutdown();

	MOD_DEC_USE_COUNT;
	return error;
}

#ifdef MODULE
int init_module(void)
{
	return devdoor_init();
}

void cleanup_module(void)
{
	if (unregister_chrdev(DOORS_MAJOR, DOORS_DEVNAME))
		printk("%s: failed to unregister chrdev.\n", doors_banner);
}
#endif /* MODULE */
