/*  Copyright (C) 1998-2002 Jason Lango
 *
 *  This file is part of the Doors for Linux Kernel Module.
 * 
 *  The Doors for Linux Kernel Module is free software; you can
 *  redistribute it and/or modify it under the terms of the GNU
 *  General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  The Doors for Linux Kernel Module is distributed in the hope that
 *  it will be useful, but WITHOUT ANY WARRANTY; without even the
 *  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *  PURPOSE.  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with the Doors for Linux Kernel Module; if not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 *  Boston, MA  02111-1307  USA
 */

/*
 * test.c - Infrastructure for Doors/Linux regression testing.
 * Jason Lango <jal@rampant.org>
 */

#include "test.h"

/*
 * Test globals.
 */
struct door_test_state *dts;

static int dts_fd = -1;
static const char *test_optstring = "hscS:C:v";
static int door = -1;
static const char *door_filename = "test";

int
test_create_door(test_server_proc_t server_proc)
{
	int fd;

	VERIFY ( door = door_create(server_proc, (void*) TEST_COOKIE, 0) );
	VERIFY ( fd = creat(test_door_filename(), 0600) );
	VERIFY ( close(fd) );
	VERIFY ( fattach(door, test_door_filename()) );

	return door;
}

void
test_destroy_door(void)
{
	if (door >= 0) {
		VERIFY ( close(door) );
		VERIFY ( fdetach(test_door_filename()) );
		VERIFY ( unlink(test_door_filename()) );
		door = -1;
	}
}

int
test_open_door(void)
{
	int fd;
	int ntries, maxtries;
	struct stat buf;

	ntries = 0;
	maxtries = 1000;

	for (;;) {
		if (ntries++ == maxtries) {
			fprintf(stderr, "Gave up waiting for door to appear "
				"as file \"%s\".\n", test_door_filename());
			exit(1);
		}

		fd = open(test_door_filename(), O_RDONLY);
		if (fd < 0) {
			if (errno != ENOENT)
				test_error("open test door: %s\n",
					strerror(errno));
		}
		else {
			/*
			 * We could possibly (though unlikely) catch
			 * it between the creat() and the fattach().
			 */
			VERIFY ( fstat(fd, &buf) );
			if (!S_ISREG(buf.st_mode))
				break;
			close(fd);
		}

		usleep(1000);
	}

	return fd;
}

const char *
test_door_filename(void)
{
	/*
	 * Note that this has to be the same for the client and
	 * server, so it can't depend on the pid.
	 */
	return door_filename;
}

void
test_error(const char *fmt, ...)
{
	va_list ap;

	fprintf(stderr, "%s: ERROR: ", dts->dts_testname);

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	test_kill_clients();
	test_kill_servers();
	exit(1);
}

void
test_verify_info(struct door_info *info, test_server_proc_t server_proc)
{
	door_attr_t	attr;
	door_attr_t	all_attr;
	pid_t		current;

	if (dts->dts_nservers &&
		info->di_proc != (door_ptr_t)(unsigned long) server_proc) {
		test_error("di_proc %lx != %lx\n",
			(unsigned long) info->di_proc,
			(unsigned long) server_proc);
	}

	if (dts->dts_nservers && !test_is_server_pid(info->di_target)) {
		test_error("di_target pid %d is not a server pid\n",
			(int) info->di_target);
	}

	if (dts->dts_nservers &&
		info->di_data != (door_ptr_t)(unsigned long) TEST_COOKIE) {
		test_error("di_data cookie %lx != %lx\n",
			(unsigned long) info->di_data,
			(unsigned long) TEST_COOKIE);
	}

	attr = info->di_attributes;
	all_attr = (DOOR_LOCAL|DOOR_UNREF|DOOR_REVOKED|DOOR_PRIVATE);

	if (attr & ~all_attr)
		test_error("door has unknown attributes\n");

	current = getpid();

	if (test_is_client_pid(current)) {
		if (attr & DOOR_LOCAL)
			test_error("client has DOOR_LOCAL set\n");
	}
	else if (test_is_server_pid(current)) {
		if (!(attr & DOOR_LOCAL))
			test_error("server doesn't have DOOR_LOCAL set\n");
	}
}

void
test_verify_cred(door_cred_t *cred)
{
	/*
	 * If we're just running as a server, we don't know anything
	 * about the client's true credentials.
	 */
	if (!dts->dts_nclients)
		return;

	if (!test_is_client_pid(cred->dc_pid))
		test_error("door_cred() returned invalid client pid %d\n",
			cred->dc_pid);

	if (cred->dc_ruid != getuid())
		test_error("door_cred() returned invalid client ruid %d\n",
			cred->dc_ruid);

	if (cred->dc_euid != geteuid())
		test_error("door_cred() returned invalid client euid %d\n",
			cred->dc_euid);
}

int
test_is_client_pid(int pid)
{
	int ii;

	for (ii = 0; ii < dts->dts_nclients; ii++)
		if (dts->dts_client_pids[ii] == pid)
			return 1;

	return 0;
}

int
test_is_server_pid(int pid)
{
	int ii;

	for (ii = 0; ii < dts->dts_nservers; ii++)
		if (dts->dts_server_pids[ii] == pid)
			return 1;

	return 0;
}

void
test_kill_servers(void)
{
	int		ii;
	pid_t		pid;

	for (ii = 0; ii < dts->dts_nservers; ii++)
		if ((pid = dts->dts_server_pids[ii]))
			kill(pid, SIGTERM);
}

void
test_kill_clients(void)
{
	int		ii;
	pid_t		pid;

	for (ii = 0; ii < dts->dts_nclients; ii++)
		if ((pid = dts->dts_client_pids[ii]))
			kill(pid, SIGTERM);
}

static void
usage(void)
{
	fprintf(stderr, "usage: %s [-h] [-s|-c] [-S <nservers>] "
		"[-C <nclients>]\n", dts->dts_testname);
}

static int
parse_options(int argc, char *argv[])
{
	int ch;

	while ((ch = getopt(argc, argv, test_optstring)) != EOF) {
		switch (ch) {
		case 's':
			dts->dts_nclients = 0;
			break;
		case 'c':
			dts->dts_nservers = 0;
			break;
		case 'S':
			dts->dts_nservers = strtoul(optarg, NULL, 10);
			break;
		case 'C':
			dts->dts_nclients = strtoul(optarg, NULL, 10);
			break;
		case 'v':
			dts->dts_verbose = 1;
			break;
		case 'h':
			usage();
			exit(0);
		case '?':
		default:
			usage();
			exit(1);
		}
	}

	return optind;
}

static void
wait_for_servers(void)
{
	int ok;
	int ii;
	int ntries, maxtries;

	maxtries = 5000;
	ntries = 0;

	for (;;) {
		ok = 1;

		if (ntries++ == maxtries) {
			fprintf(stderr, "Waited %d times for startup... "
				"bailing out.\n", maxtries);
			abort();
		}

		for (ii = 0; ii < dts->dts_nservers; ii++)
			if (!dts->dts_server_ready[ii])
				ok = 0;

		if (ok)
			break;
		else
			usleep(1000);
	}
}

static void
init_dts(void)
{
	char dts_filename[256];
	void *addr;

	/*
	 * Set up a region of shared memory for the "door test state"
	 * structure.
	 */

	if (dts_fd < 0) {
		sprintf(dts_filename, "door_test_state.%d", getpid());
		VERIFY ( dts_fd = open(dts_filename, O_RDWR|O_TRUNC|O_CREAT,
			0600) );
		VERIFY ( unlink(dts_filename) );
		VERIFY ( ftruncate(dts_fd, sizeof(*dts)) );
	}

	addr = mmap(0, sizeof(*dts), PROT_READ|PROT_WRITE, MAP_SHARED,
		dts_fd, 0);
	if (addr == (void*)-1) {
		perror("ERROR: mmap of dts file");
		abort();
	}

	dts = (struct door_test_state*) addr;
}

static int
client_start(int n, int argc, char *argv[])
{
	init_dts();
	VERIFY ( dts->dts_client_pids[n] = getpid() );
	wait_for_servers();
	return client_main(n, argc, argv);
}

static int
server_start(int n, int argc, char *argv[])
{
	int error;

	init_dts();
	VERIFY ( dts->dts_server_pids[n] = getpid() );

	error = server_main(n, argc, argv);
	dts->dts_server_ready[n] = 1;
	if (error)
		return error;

	for (;;)
		pause();
	return -1;
}

static void
process_status_error(int pid, int status)
{
	if (WIFEXITED(status))
		test_error("pid %d exited with status %d\n", pid,
			WEXITSTATUS(status));
	else if (WIFSIGNALED(status))
		test_error("pid %d exited with signal %d\n", pid,
			WTERMSIG(status));
	else if (WIFSTOPPED(status))
		test_error("pid %d stopped with signal %d\n", pid,
			WSTOPSIG(status));
	else
		test_error("pid %d exited with unknown status %d\n", pid,
			status);
}

static void
process_end(int pid, int status)
{
	if (test_is_client_pid(pid))
		--dts->dts_nclients_remain;
	else if (test_is_server_pid(pid))
		--dts->dts_nservers_remain;
	else
		test_error("child pid %d neither client nor server\n", pid);

	if (status) {
		process_status_error(pid, status);
		test_kill_clients();
		test_kill_servers();
	}

	if (!dts->dts_nclients_remain)
		test_kill_servers();
}

int
main(int argc, char *argv[])
{
	int ii;
	int pid;
	int retval;

	init_dts();

	dts->dts_testname = strrchr(argv[0], '/');
	if (dts->dts_testname)
		++dts->dts_testname;
	else
		dts->dts_testname = argv[0];

	dts->dts_nclients = 1;
	dts->dts_nservers = 1;
	dts->dts_nclients_remain = 0;
	dts->dts_nservers_remain = 0;

	ii = parse_options(argc, argv);
	argc -= ii;
	argv += ii;

	printf("%s: Testing %s.\n", dts->dts_testname, test_description());

	/*
	 * Start the client and server processes.
	 */

	if (dts->dts_nservers) {
		for (ii = 0; ii < dts->dts_nservers; ii++) {
			VERIFY ( pid = fork() );
			if (!pid)
				return server_start(ii, argc, argv);
			++dts->dts_nservers_remain;
		}
	}

	if (dts->dts_nclients) {
		for (ii = 0; ii < dts->dts_nclients; ii++) {
			VERIFY ( pid = fork() );
			if (!pid)
				return client_start(ii, argc, argv);
			++dts->dts_nclients_remain;
		}
	}

	/*
	 * If any of the child processes exit with an error value, return that
	 * value.
	 */

	retval = 0;
	for (;;) {
		int status;

		pid = wait(&status);
		if (pid < 0) {
			if (errno == ECHILD)
				break;
			else {
				perror("ERROR: wait");
				abort();
			}
		}

		/* Special case: killing a server */
		if (test_is_server_pid(pid) && WIFSIGNALED(status))
			status = 0;

		if (status)
			retval = status;

		process_end(pid, status);
	}

	if (!retval)
		printf("%s: Passed.\n", dts->dts_testname);
	else
		printf("%s: *** TEST FAILED! ***\n", dts->dts_testname);

	return retval;
}
